import type { AxiosInstance, AxiosRequestConfig } from 'axios'
import qs from 'qs'
import { ContentTypeEnum, app } from './constant'
import { defaultTransform, defaultTransformConfig } from './transformer'

import type { BaseApi } from './typing'
import { checkContentType, checkIsGet, checkParams, isString, runFunction } from './utils'
type BaseApiNoMethod<D> = Omit<BaseApi<D>, 'method'>

interface UploadFileParams<D>{
  data?: D
  name?: string
  file: File | Blob
  filename?: string
  [key: string]: any
}

export interface InitHttpCallback {
  request <R=any, D=any>(config: BaseApi<D>): Promise<R>
  get <R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R>
  post <R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R>
  put <R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R>
  form <R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R>
  upload <R=any>(config: BaseApiNoMethod<FormData>): Promise<R>
  delete <R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R>
  uploadFile <R=any, D=any>(config: Omit<BaseApiNoMethod<any>, 'params'|'data'>, params: UploadFileParams<D>): Promise<R>
}

export default function initHttp(instance: AxiosInstance): InitHttpCallback {
  // 请求request
  async function request<R = any, D=any>(config: BaseApi<D>): Promise<R> {
    const { beforeRequest, afterRequest, beforeRequestConfig, beforeRequestQuery, onError, responseTransform, urlParser, requestTransform, dataType, ...restConfig } = config
    let newConfig
    // 处理请求类型
    if (dataType) {
      switch (dataType) {
        case 'form':
          restConfig.headers = {
            ...restConfig.headers,
            'Content-type': ContentTypeEnum.FORM,
          }
          break
        case 'upload':
          restConfig.headers = {
            ...restConfig.headers,
            'Content-type': ContentTypeEnum.UPLOAD,
          }
          break
        case 'json':
          restConfig.headers = {
            ...restConfig.headers,
            'Content-type': ContentTypeEnum.JSON,
          }
          break
        default:
          if (!checkContentType(restConfig.headers as any)) {
            restConfig.headers = {
              ...restConfig.headers,
              'Content-type': ContentTypeEnum.JSON,
            }
          }
      }
    }
    const requestFun = requestTransform || app.hooks?.requestTransform || defaultTransformConfig
    newConfig = checkParams(requestFun(restConfig), restConfig)
    newConfig = checkParams(await runFunction<AxiosRequestConfig<D>>(app.hooks?.beforeRequestConfig, newConfig), newConfig)
    newConfig = checkParams(await runFunction<AxiosRequestConfig<D>>(beforeRequestConfig, newConfig), newConfig)
    // 拦截url的处理钩子
    newConfig.url = checkParams(await runFunction(app.hooks?.urlParser, newConfig.url), newConfig.url)
    newConfig.url = checkParams(await runFunction(urlParser, newConfig.url), newConfig.url)
    // GET 请求钩子特殊处理
    if (checkIsGet(newConfig?.method)) {
      // 全局的钩子
      newConfig.params = checkParams(await runFunction(app.hooks?.beforeRequest, newConfig.params), newConfig.params)
      // 单个请求的钩子
      newConfig.params = checkParams(await runFunction(beforeRequest, newConfig.params), newConfig.params)
    }
    else {
      // 管他存不存在的都要走这些钩子
      newConfig.data = checkParams(await runFunction(app.hooks?.beforeRequest, newConfig.data), newConfig.data)
      newConfig.data = checkParams(await runFunction(beforeRequest, newConfig.data), newConfig.data)
      // 如果存在query的参数，就直接使用
      newConfig.params = checkParams(await runFunction(app.hooks?.beforeRequestQuery, newConfig.params), newConfig.params)
      newConfig.params = checkParams(await runFunction(beforeRequestQuery, newConfig.params), newConfig.params)
    }

    try {
      let response = await instance.request<R>(newConfig)
      // 请求后的钩子函数
      response = await checkParams(await runFunction(afterRequest, response), response)
      // transform优先级从左到右，由大到小： 接口自定义 -> 全局自定义 -> 默认
      const responseFun = responseTransform || app.hooks?.responseTransform || defaultTransform
      // 这里还可以加处理
      return responseFun(response)
    }
    catch (e) {
      app.hooks?.onError?.(e)
      // TODO
      onError?.(e)
      throw e
    }
  }

  // get请求
  async function get<R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R> {
    return request<R, D>({ ...config, method: 'GET' })
  }
  // POST请求
  async function post<R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R> {
    return request<R, D>({ ...config, method: 'POST' })
  }
  // Put请求
  async function put<R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R> {
    return request<R, D>({ ...config, method: 'PUT' })
  }
  // delete请求
  async function deleteFun<R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R> {
    return request<R, D>({ ...config, method: 'DELETE' })
  }
  // upload请求
  async function upload<R=any>(config: BaseApiNoMethod<FormData>): Promise<R> {
    return request<R>({
      ...config,
      method: 'POST',
      headers: {
        ...config.headers,
        'Content-Type': ContentTypeEnum.UPLOAD,
      },
    })
  }

  async function uploadFile<R=any, D = any>(config: Omit<BaseApiNoMethod<any>, 'params'|'data'>, params: UploadFileParams<D>): Promise<R> {
    // 处理上传文件类型
    const formData = new FormData()
    if (params.data) {
      for (const pKey in params.data) {
        const value = params.data[pKey]
        if (Array.isArray(value)) {
          value.forEach((item) => {
            formData.append(`${pKey}[]`, item)
          })
        }
        else {
          formData.append(pKey, (value as any))
        }
      }
    }
    formData.append(params.name || 'file', params.file, params.filename)
    if (params.filename)
      formData.append('fileName', params.filename)

    return upload<R>(
      {
        ...config,
        data: formData,
      },
    )
  }

  // form请求
  async function form<R=any, D=any>(config: BaseApiNoMethod<D>): Promise<R> {
    return request<R, string>({
      ...config,
      method: 'POST',
      headers: {
        ...config.headers,
        'Content-Type': ContentTypeEnum.FORM,
      },
      data: isString(config.data) ? config.data : qs.stringify(config.data),
    })
  }
  return {
    request,
    get,
    delete: deleteFun,
    post,
    upload,
    form,
    put,
    uploadFile,
  }
}
